#include "../../internal-include/BootLoader/boot_loader.hpp"
#include "../../internal-include/GPU/gpu_internal.hpp"
#include <stdio.h>

#include <PSX/GTE/gte.hpp>
#include <PSX/System/syscalls.hpp>
#include "../../reference/inline_n.h"
#include <math.h>

extern "C" uint32_t __heap_start;
extern "C" uint32_t __bss_start;
extern "C" uint32_t __bss_end;
extern "C" uint32_t __bss_len;
extern "C" uint32_t __planschi_start;
extern "C" uint32_t __planschi_end;

namespace JabyEngine {
    static void test_bios_font() {
        static constexpr uint16_t SJIS = 0x83B5;

        const auto*      font     = SysCall::Krom2RawAdd(SJIS);
        const auto*const font_end = font + 16;
        printf("Loading SJIS from @0x%p\n", font);
        
        for(; font < font_end; font++) {
            const auto cur_char = __builtin_bswap16(*font);
            for(uint16_t shift = 1 << 15; shift; shift = shift >> 1) {
                printf(cur_char & shift ? "#" : " ");
            }

            printf("\n");
        }
    }

    static void test_gte() {
#define gte_MulMatrix0(r1,r2,r3)				\
                {	gte_SetRotMatrix(r1);	\
                    gte_ldclmv(r2);		\
                    gte_rtir();		\
                    gte_stclmv(r3);		\
                    gte_ldclmv((char*)r2+2);\
                    gte_rtir();		\
                    gte_stclmv((char*)r3+2);\
                    gte_ldclmv((char*)r2+4);\
                    gte_rtir();		\
                    gte_stclmv((char*)r3+4);	}
                    
        const auto m0 = GTE::MATRIX::identity();
        const auto m1 = GTE::MATRIX::identity();
        auto       m2 = GTE::MATRIX::identity();

        asm("# PLANSCHI START");
        JabyEngine::GTE::multiply_matrix(m0, m1, m2);
        asm("# PLANSCHI END");

        #define PI 3.14159265 // v 90°???

        static const auto display_stuff = [](int32_t angle) {
            static const auto wanna_be_float = [](int32_t value) -> int32_t {
                return (value*10000)/(4096);
            };

            const auto sin_value = sin(GTE::in_degree(angle));
            const auto cos_value = cos(GTE::in_degree(angle));
            
            printf("Sine of %i = %i (%i)\n", angle, sin_value, wanna_be_float(sin_value));
            printf("Cos  of %i = %i (%i)\n", angle, cos_value, wanna_be_float(cos_value));
        };

        static const auto print_matrix = [](const GTE::MATRIX& matrix) {
            printf("{%i, %i, %i}\n", matrix.rot[0][0], matrix.rot[0][1], matrix.rot[0][2]);
            printf("{%i, %i, %i}\n", matrix.rot[1][0], matrix.rot[1][1], matrix.rot[1][2]);
            printf("{%i, %i, %i}\n", matrix.rot[2][0], matrix.rot[2][1], matrix.rot[2][2]);
            printf("{%i, %i, %i}\n", matrix.trans[0],  matrix.trans[1],  matrix.trans[2]);
        };

        static const auto make_matrix = [](int16_t mult) -> GTE::MATRIX {
            return GTE::MATRIX {
                .rot   = {
                    {static_cast<int16_t>(1*mult), static_cast<int16_t>(2*mult), static_cast<int16_t>(3*mult)},
                    {static_cast<int16_t>(2*mult), static_cast<int16_t>(3*mult), static_cast<int16_t>(1*mult)},
                    {static_cast<int16_t>(3*mult), static_cast<int16_t>(1*mult), static_cast<int16_t>(2*mult)},
                },
                .trans = {6*mult, 7*mult, 8*mult}
            };
        };

        display_stuff(25);
        display_stuff(45);
        display_stuff(75);
        display_stuff(90);

        for(size_t n = 0; n < 3; n++) {
            GTE::push_matrix_and_set(make_matrix(n + 1));
        }

        for(size_t n = 0; n < 3; n++) {
            print_matrix(GTE::get_and_pop_matrix());
            printf("---\n");
        }
    }

    namespace boot {
        namespace Start { 
            static void setup() {
                static constexpr auto DebugX     = 1;
                static constexpr auto DebugY     = 0;
                static constexpr auto DebugScale = 1.0;

                BIOS::identify();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::Grey(), DebugX, DebugY, DebugScale);
                DMA::setup();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::White(), DebugX, DebugY, DebugScale);
                Periphery::setup();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::Red(), DebugX, DebugY, DebugScale);
                SPU::stop_voices();
                
                __debug_boot_color_at(::JabyEngine::GPU::Color24::Green(), DebugX, DebugY, DebugScale);
                CD::setup();
                __debug_boot_color_at(::JabyEngine::GPU::Color24::Blue(), DebugX, DebugY, DebugScale);
                Timer::setup();

                __debug_boot_color_at(::JabyEngine::GPU::Color24::Yellow(), DebugX, DebugY, DebugScale);
                GPU::setup();
                GPU::display_logo();
                GTE::setup();
                test_bios_font();
                test_gte();

                //Pause??
                SPU::setup();
            }
        }
    }

    void start() {
        static constexpr auto DebugX     = 0;
        static constexpr auto DebugY     = 0;
        static constexpr auto DebugScale = 1.0;

        __debug_boot_print_at(GPU::Color24::White(), DebugX, DebugY, DebugScale, "Starting Planschbecken\n");
        printf("Heap starts @0x%p\n", &__heap_start);
        printf("BSS      from 0x%p to 0x%p (%u)\n", &__bss_start, &__bss_end, __bss_len);
        __debug_boot_print_at(GPU::Color24::Green(), DebugX, DebugY, DebugScale, "PLANSCHI from 0x%p to 0x%p\n", &__planschi_start, &__planschi_end);
        boot::Start::setup();
        
        __debug_boot_print_at(GPU::Color24::Red(), DebugX, DebugY, DebugScale, "Running main...\n");
        GPU::internal::wait_vsync(0);
        run();
    }
}